# -*- coding: utf-8 -*-
from django.contrib import admin

from .conf import settings
from .models import Cookie, CookieGroup, LogItem


class CookieAdmin(admin.ModelAdmin):
    list_display = ("varname", "name", "cookiegroup", "path", "domain", "get_version")
    search_fields = ("name", "domain", "cookiegroup__varname", "cookiegroup__name")
    readonly_fields = ("varname",)
    list_filter = ("cookiegroup",)


class CookieGroupAdmin(admin.ModelAdmin):
    list_display = ("varname", "name", "is_required", "is_deletable", "get_version")
    search_fields = (
        "varname",
        "name",
    )
    list_filter = (
        "is_required",
        "is_deletable",
    )


class LogItemAdmin(admin.ModelAdmin):
    list_display = ("action", "cookiegroup", "version", "created")
    list_filter = ("action", "cookiegroup")
    readonly_fields = ("action", "cookiegroup", "version", "created")
    date_hierarchy = "created"


admin.site.register(Cookie, CookieAdmin)
admin.site.register(CookieGroup, CookieGroupAdmin)
if settings.COOKIE_CONSENT_LOG_ENABLED:
    admin.site.register(LogItem, LogItemAdmin)



from django.test import TestCase, override_settings
from django.test.client import RequestFactory
from django.urls import reverse

from cookie_consent.cache import delete_cache
from cookie_consent.models import Cookie, CookieGroup

factory = RequestFactory()


@override_settings(COOKIE_CONSENT_OPT_OUT=False)
class CleanCookiesMiddlewareTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()

        cls.cookie_group = CookieGroup.objects.create(
            varname="optional",
            name="Optional (test) cookies",
        )
        cls.cookie = Cookie.objects.create(
            cookiegroup=cls.cookie_group,
            name="optional_test_cookie",
            domain="127.0.0.1",
            path="/",
        )

    def setUp(self):
        super().setUp()
        self.addCleanup(delete_cache)

    def _accept_and_set_cookie(self):
        with self.subTest("initial setup"):
            # ensure we start with cookies first being accepted
            endpoint = reverse("cookie_consent_accept", kwargs={"varname": "optional"})

            consent_response = self.client.post(
                endpoint,
                follow=True,
                HTTP_X_REQUESTED_WITH="XMLHttpRequest",
            )
            self.assertEqual(consent_response.status_code, 200)

            # hit the test page to set the cookie
            self.client.get(reverse("test_page"))
            self.assertIn("optional_test_cookie", self.client.cookies)

    def assertCookieDeleted(self, name: str):
        if name not in self.client.cookies:
            self.fail(
                "Cookie not present in client cookies, which is required to delete it"
            )

        # deleting a cookie is done by setting the expiry date to the past/max-age 0
        # so that the browser effectively is instructed to delete the cookie
        cookie = self.client.cookies[name]
        cookie_dict = dict(cookie)
        self.assertEqual(cookie_dict["max-age"], 0)
        self.assertIn("1970", cookie_dict["expires"])
        self.assertEqual(cookie.value, "")

    def test_middleware_decline_previously_accepted_cookiegroup_cookies_are_deleted(
        self,
    ):
        self._accept_and_set_cookie()

        with self.subTest("decline prevously accepted group"):
            url = reverse("cookie_consent_decline", kwargs={"varname": "optional"})

            decline_response = self.client.post(url, follow=True)

            self.assertEqual(decline_response.status_code, 200)

        # fetch the test page and assert the middleware deleted the cookie
        self.client.get(reverse("test_page"))

        self.assertCookieDeleted("optional_test_cookie")

    def test_middleware_no_cookie_consent_cookie_present_cookies_are_deleted(self):
        self._accept_and_set_cookie()
        # Delete cookie_consent cookie
        del self.client.cookies["cookie_consent"]

        # fetch the test page and assert the middleware deleted the cookie
        self.client.get(reverse("test_page"))

        # Check if cookie_consent cookie is deleted
        self.assertCookieDeleted("optional_test_cookie")

    def test_cookie_consent_disabled(self):
        self._accept_and_set_cookie()

        with override_settings(COOKIE_CONSENT_ENABLED=False):
            self.client.get(reverse("test_page"))

        cookie = self.client.cookies["optional_test_cookie"]
        self.assertEqual(cookie.value, "optional cookie set from django")

    def test_cookie_group_not_deletable(self):
        self.cookie_group.is_deletable = False
        self.cookie_group.save()
        self._accept_and_set_cookie()

        self.client.get(reverse("test_page"))

        cookie = self.client.cookies["optional_test_cookie"]
        self.assertEqual(cookie.value, "optional cookie set from django")

    @override_settings(COOKIE_CONSENT_OPT_OUT=True)
    def test_with_opt_out_behaviour(self):
        # set the cookie
        self.client.get(reverse("test_page"), {"force": "1"})

        # call view to run middleware
        self.client.get(reverse("test_page"))

        cookie = self.client.cookies["optional_test_cookie"]
        self.assertEqual(cookie.value, "optional cookie set from django")



# -*- coding: utf-8 -*-
from django.test import TestCase, override_settings

from cookie_consent.cache import delete_cache, get_cookie, get_cookie_group
from cookie_consent.models import Cookie, CookieGroup


class CacheTest(TestCase):
    def setUp(self):
        super().setUp()
        self.addCleanup(delete_cache)

        self.cookie_group = CookieGroup.objects.create(
            varname="optional",
            name="Optional",
        )
        self.cookie = Cookie.objects.create(
            cookiegroup=self.cookie_group,
            name="foo",
        )

    def test_get_cookie_group(self):
        self.assertEqual(get_cookie_group("optional"), self.cookie_group)

    def test_get_cookie(self):
        cookie_group = get_cookie_group("optional")
        self.assertEqual(get_cookie(cookie_group, "foo", ""), self.cookie)

    def test_caching(self):
        CookieGroup.objects.create(
            varname="foo",
            name="Foo",
        )
        with self.assertNumQueries(2):
            cookie_group = get_cookie_group("optional")
            get_cookie_group("foo")
            get_cookie(cookie_group, "foo", "")

    def test_caching_expire(self):
        with self.assertNumQueries(2):
            cookie_group = get_cookie_group("optional")

        self.cookie_group.name = "Bar"
        self.cookie_group.save()

        with self.assertNumQueries(2):
            cookie_group = get_cookie_group("optional")
        self.assertEqual(cookie_group.name, "Bar")

    @override_settings(
        CACHES={"tests": {"BACKEND": "django.core.cache.backends.dummy.DummyCache"}},
        COOKIE_CONSENT_CACHE_BACKEND="tests",
    )
    def test_can_override_cache_settings(self):
        """
        Assert that the cache backend/settings can be swapped out in tests.

        Regression test for #41
        """
        CookieGroup.objects.create(
            varname="foo",
            name="Foo",
        )
        # expect multiple calls to not be cached because of the no-op cache
        with self.assertNumQueries(2 + 2):
            get_cookie_group("optional")
            get_cookie_group("foo")



# -*- coding: utf-8 -*-
import re
from typing import TypedDict

from django.core.validators import RegexValidator
from django.db import models
from django.utils.translation import gettext_lazy as _

from cookie_consent.cache import delete_cache

COOKIE_NAME_RE = re.compile(r"^[-_a-zA-Z0-9]+$")
validate_cookie_name = RegexValidator(
    COOKIE_NAME_RE,
    _(
        "Enter a valid 'varname' consisting of letters, numbers"
        ", underscores or hyphens."
    ),
    "invalid",
)


class CookieGroupDict(TypedDict):
    varname: str
    name: str
    description: str
    is_required: bool
    # TODO: should we output this? page cache busting would be
    # required if we do this. Alternatively, set up a JSONView to output these?
    # version: str


class CookieGroup(models.Model):
    varname = models.CharField(
        _("Variable name"), max_length=32, validators=[validate_cookie_name]
    )
    name = models.CharField(_("Name"), max_length=100, blank=True)
    description = models.TextField(_("Description"), blank=True)
    is_required = models.BooleanField(
        _("Is required"),
        help_text=_("Are cookies in this group required."),
        default=False,
    )
    is_deletable = models.BooleanField(
        _("Is deletable?"),
        help_text=_("Can cookies in this group be deleted."),
        default=True,
    )
    ordering = models.IntegerField(_("Ordering"), default=0)
    created = models.DateTimeField(_("Created"), auto_now_add=True, blank=True)

    class Meta:
        verbose_name = _("Cookie Group")
        verbose_name_plural = _("Cookie Groups")
        ordering = ["ordering"]

    def __str__(self):
        return self.name

    def get_version(self) -> str:
        try:
            return str(self.cookie_set.all()[0].get_version())
        except IndexError:
            return ""

    def delete(self, *args, **kwargs):
        super(CookieGroup, self).delete(*args, **kwargs)
        delete_cache()

    def save(self, *args, **kwargs):
        super(CookieGroup, self).save(*args, **kwargs)
        delete_cache()

    def for_json(self) -> CookieGroupDict:
        return {
            "varname": self.varname,
            "name": self.name,
            "description": self.description,
            "is_required": self.is_required,
            # "version": self.get_version(),
        }


class Cookie(models.Model):
    cookiegroup = models.ForeignKey(
        CookieGroup,
        verbose_name=CookieGroup._meta.verbose_name,
        on_delete=models.CASCADE,
    )
    name = models.CharField(_("Name"), max_length=250)
    description = models.TextField(_("Description"), blank=True)
    path = models.TextField(_("Path"), blank=True, default="/")
    domain = models.CharField(_("Domain"), max_length=250, blank=True)
    created = models.DateTimeField(_("Created"), auto_now_add=True, blank=True)

    class Meta:
        verbose_name = _("Cookie")
        verbose_name_plural = _("Cookies")
        ordering = ["-created"]

    def __str__(self):
        return "%s %s%s" % (self.name, self.domain, self.path)

    @property
    def varname(self):
        return "%s=%s:%s" % (self.cookiegroup.varname, self.name, self.domain)

    def get_version(self):
        return self.created.isoformat()

    def delete(self, *args, **kwargs):
        super(Cookie, self).delete(*args, **kwargs)
        delete_cache()

    def save(self, *args, **kwargs):
        super(Cookie, self).save(*args, **kwargs)
        delete_cache()


ACTION_ACCEPTED = 1
ACTION_DECLINED = -1
ACTION_CHOICES = (
    (ACTION_DECLINED, _("Declined")),
    (ACTION_ACCEPTED, _("Accepted")),
)


class LogItem(models.Model):
    action = models.IntegerField(_("Action"), choices=ACTION_CHOICES)
    cookiegroup = models.ForeignKey(
        CookieGroup,
        verbose_name=CookieGroup._meta.verbose_name,
        on_delete=models.CASCADE,
    )
    version = models.CharField(_("Version"), max_length=32)
    created = models.DateTimeField(_("Created"), auto_now_add=True, blank=True)

    def __str__(self):
        return "%s %s" % (self.cookiegroup.name, self.version)

    class Meta:
        verbose_name = _("Log item")
        verbose_name_plural = _("Log items")
        ordering = ["-created"]



from textwrap import dedent
from typing import Any, Dict, Optional

from django.template import Context, Template

import pytest


def render(tpl: str, context: Optional[Dict[str, Any]] = None) -> str:
    template = Template(dedent(tpl).strip())
    return template.render(Context(context))


NOT_ACCEPT_OR_DECLINED_TEMPLATE = """
{% load cookie_consent_tags %}
{% not_accepted_or_declined_cookie_groups request as cookie_groups %}
{% if cookie_groups %}FOUND COOKIES{% else %}NO COOKIES{% endif %}
"""


@pytest.mark.django_db
def test_not_accepted_or_declined_cookie_groups_only_required_cookies(
    required_cookiegroup, rf
):
    context = {"request": rf.get("/")}

    output = render(NOT_ACCEPT_OR_DECLINED_TEMPLATE, context).strip()

    assert output == "NO COOKIES"


@pytest.mark.django_db
def test_not_accepted_or_declined_cookie_groups_only_optional_cookies(
    optional_cookiegroup, rf
):
    context = {"request": rf.get("/")}

    output = render(NOT_ACCEPT_OR_DECLINED_TEMPLATE, context).strip()

    assert output == "FOUND COOKIES"


def test_not_accepted_or_declined_cookie_groups_required_and_optional_cookies(
    required_cookiegroup, optional_cookiegroup, rf
):
    context = {"request": rf.get("/")}

    output = render(NOT_ACCEPT_OR_DECLINED_TEMPLATE, context).strip()

    assert output == "FOUND COOKIES"



# -*- coding: utf-8 -*-
from django.core.cache import caches

from cookie_consent.conf import settings

CACHE_KEY = "cookie_consent_cache"
CACHE_TIMEOUT = 60 * 60


def _get_cache():
    """
    Lazily wrap around django.core.cache.

    This prevents the cache object to be resolved at import-time, which breaks the
    `django.test.override_settings` functionality for projects adding tests for this
    package, see https://github.com/bmihelac/django-cookie-consent/issues/41.
    """
    return caches[settings.COOKIE_CONSENT_CACHE_BACKEND]


def delete_cache():
    cache = _get_cache()
    cache.delete(CACHE_KEY)


def all_cookie_groups():
    cache = _get_cache()
    items = cache.get(CACHE_KEY)
    if items is None:
        from cookie_consent.models import CookieGroup

        qs = CookieGroup.objects.filter(is_required=False)
        qs = qs.prefetch_related("cookie_set")
        # items = qs.in_bulk(field_name="varname")
        # FIXME -> doesn't work because varname is not a unique fieldl, we need to
        # make this unique
        items = {group.varname: group for group in qs}
        cache.set(CACHE_KEY, items, CACHE_TIMEOUT)
    return items


def get_cookie_group(varname):
    return all_cookie_groups().get(varname)


def get_cookie(cookie_group, name, domain):
    for cookie in cookie_group.cookie_set.all():
        if cookie.name == name and cookie.domain == domain:
            return cookie
    return None



# -*- coding: utf-8 -*-
from datetime import datetime

from django.test import TestCase
from django.test.client import RequestFactory
from django.test.utils import override_settings

from cookie_consent.conf import settings
from cookie_consent.models import Cookie, CookieGroup
from cookie_consent.util import (
    dict_to_cookie_str,
    get_accepted_cookies,
    get_cookie_groups,
    get_cookie_value_from_request,
    is_cookie_consent_enabled,
    parse_cookie_str,
)


class UtilTest(TestCase):
    def setUp(self):
        self.cookie_group = CookieGroup.objects.create(
            varname="optional",
            name="Optional",
        )
        self.cookie = Cookie.objects.create(
            cookiegroup=self.cookie_group,
            name="foo",
        )
        self.factory = RequestFactory()
        self.request = self.factory.get("")

    def test_parse_cookie_str(self):
        cookie_str = "foo=2013-06-04T01:08:58.262162|bar=2013-06-04T01:08:58"
        res = parse_cookie_str(cookie_str)
        dic = {
            "foo": "2013-06-04T01:08:58.262162",
            "bar": "2013-06-04T01:08:58",
        }
        self.assertEqual(res, dic)

    def test_dict_to_cookie_str(self):
        cookie_str = "|"
        dic = {
            "foo": "2013-06-04T01:08:58.262162",
            "bar": "2013-06-04T01:08:58",
        }
        cookie_str = dict_to_cookie_str(dic)
        self.assertEqual(parse_cookie_str(cookie_str), dic)

    def test_get_cookie_value_from_request(self):
        cookie_str = dict_to_cookie_str({"optional": self.cookie_group.get_version()})
        self.request.COOKIES[settings.COOKIE_CONSENT_NAME] = cookie_str
        res = get_cookie_value_from_request(self.request, "optional")
        self.assertTrue(res)

    def test_get_cookie_value_from_request_declined(self):
        cookie_str = dict_to_cookie_str({"optional": datetime(1999, 1, 1).isoformat()})
        self.request.COOKIES[settings.COOKIE_CONSENT_NAME] = cookie_str
        res = get_cookie_value_from_request(self.request, "optional")
        self.assertFalse(res)

    def test_get_cookie_value_from_request_empty(self):
        res = get_cookie_value_from_request(self.request, "optional")
        self.assertIsNone(res)

    def test_get_cookie_value_from_request_added_cookies(self):
        cookie_str = dict_to_cookie_str(
            {
                "optional": self.cookie_group.get_version(),
            }
        )
        Cookie.objects.create(
            cookiegroup=self.cookie_group,
            name="bar",
            domain=".example.com",
        )
        self.request.COOKIES[settings.COOKIE_CONSENT_NAME] = cookie_str
        res = get_cookie_value_from_request(self.request, "optional")
        self.assertIsNone(res)

    def test_get_cookie_value_from_request_specific_cookie(self):
        cookie_str = dict_to_cookie_str({"optional": self.cookie_group.get_version()})
        self.request.COOKIES[settings.COOKIE_CONSENT_NAME] = cookie_str
        res = get_cookie_value_from_request(self.request, "optional", "foo:")
        self.assertTrue(res)

        Cookie.objects.create(
            cookiegroup=self.cookie_group,
            name="bar",
            domain=".example.com",
        )
        res = get_cookie_value_from_request(
            self.request, "optional", "bar:.example.com"
        )
        self.assertFalse(res)

        res = get_cookie_value_from_request(self.request, "optional", "foo:")
        self.assertTrue(res)

        cookie_str = dict_to_cookie_str({"optional": self.cookie_group.get_version()})
        self.request.COOKIES[settings.COOKIE_CONSENT_NAME] = cookie_str
        res = get_cookie_value_from_request(
            self.request, "optional", "bar:.example.com"
        )
        self.assertTrue(res)

    def test_is_cookie_consent_enabled(self):
        self.assertTrue(is_cookie_consent_enabled(None))

    @override_settings(COOKIE_CONSENT_ENABLED=lambda r: False)
    def test_is_cookie_consent_enabled_callable(self):
        self.assertFalse(is_cookie_consent_enabled(None))

    def test_get_cookie_groups(self):
        self.assertIn(self.cookie_group, get_cookie_groups("optional"))

        cookie_group2 = CookieGroup.objects.create(
            varname="foo",
            name="foo",
        )
        self.assertIn(self.cookie_group, get_cookie_groups("foo,optional"))
        self.assertIn(cookie_group2, get_cookie_groups("foo,optional"))

    def test_get_accepted_cookies(self):
        cookie_str = dict_to_cookie_str({"optional": self.cookie_group.get_version()})
        self.request.COOKIES[settings.COOKIE_CONSENT_NAME] = cookie_str
        cookies = get_accepted_cookies(self.request)
        self.assertIn(self.cookie, cookies)



# -*- coding: utf-8 -*-
from typing import Optional

from .cache import all_cookie_groups
from .conf import settings
from .util import get_cookie_dict_from_request, is_cookie_consent_enabled


def _should_delete_cookie(group_version: Optional[str]) -> bool:
    # declined after it was accepted (and set) before
    if group_version == settings.COOKIE_CONSENT_DECLINE:
        return True

    # if you need to opt-out instead of opt-in, then we only delete the cookie in the
    # above scenario -> when the group is explicitly declined
    if settings.COOKIE_CONSENT_OPT_OUT:
        return False

    # when we are opt-in and have no information whether the cookie group was accepted
    # or declined, delete the cookie(s).
    if group_version is None:
        return True

    return False


class CleanCookiesMiddleware:
    """
    Clean declined or non-accepted cookies.

    Note that this only applies if COOKIE_CONSENT_OPT_OUT is not set.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        if is_cookie_consent_enabled(request):
            self.process_response(request, response)
        return response

    def process_response(self, request, response):
        cookie_dic = get_cookie_dict_from_request(request)

        cookies_to_delete = []
        for cookie_group in all_cookie_groups().values():
            if not cookie_group.is_deletable:
                continue

            group_version = cookie_dic.get(cookie_group.varname, None)
            for cookie in cookie_group.cookie_set.all():
                if cookie.name not in request.COOKIES:
                    continue
                if _should_delete_cookie(group_version):
                    cookies_to_delete.append(cookie)

        for cookie in cookies_to_delete:
            response.delete_cookie(cookie.name, cookie.path, cookie.domain)

        return response



try:
    from django.utils.http import url_has_allowed_host_and_scheme
except ImportError:  # django < 3.0
    from django.utils.http import is_safe_url as url_has_allowed_host_and_scheme

try:  # django >= 4.1
    from django.contrib.auth.views import RedirectURLMixin
except ImportError:
    from django.contrib.auth.views import (
        SuccessURLAllowedHostsMixin as RedirectURLMixin,
    )

__all__ = ["url_has_allowed_host_and_scheme", "RedirectURLMixin"]
